#
# Copyright 2011 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#

import os, sys
import errno
import signal
import time
import posixpath

ourname = os.path.basename(sys.argv[0])

EXIT_OK, EXIT_ERROR, EXIT_USAGE = range(3)

class Error(Exception): pass

def load_app():
    try:
        if not os.path.isfile('app.py'):
            raise ImportError
        sys.path.append(os.getcwd())
        from app import app
    except ImportError:
        raise Error('current directory does not contain an Albatross '
                    'application (app.py not found, or not able to import app '
                    'and cfg objects from app.py).')
    return app


def split_posix_path(path):
    parts = []
    path, part = posixpath.split(path)
    while path:
        parts.append(part)
        path, part = os.path.split(path)
    parts.append(part)
    parts.reverse()
    return parts


_skeleton_dir = None

def get_skeleton_dir():
    global _skeleton_dir
    if _skeleton_dir is None:
        thisdir = os.path.dirname(__file__)
        search = [
            os.path.abspath(os.path.join(thisdir, os.pardir, os.pardir, 'skeleton')),
            os.path.join(sys.prefix, 'share', 'albatross', 'skeleton'),
            os.path.join(sys.prefix, 'local', 'share', 'albatross', 'skeleton'),
        ]
        for skeldir in search:
            if os.path.exists(skeldir):
                _skeleton_dir = skeldir
                break
        else:
            raise Error('Application skeleton not found')
    return _skeleton_dir


class Pipe:

    """
    A simple wrapper around a unix pipe, intended to be used to
    synchronise parent and child processes after a fork(2).

    One process holds the read end, the other process holds the write end.
    The first use decides which direction the pipe runs in, with the
    unused file descriptors being closed (so the other end can detect
    EOF).

    On destruction of the object, any open file descriptors are closed.

    This is Unix-specific.
    """

    def __init__(self):
        self.read_fd, self.write_fd = os.pipe()

    def write(self, buf):
        if self.read_fd >= 0:
            try:
                os.close(self.read_fd)
            except:
                pass
            self.read_fd = -1
        return os.write(self.write_fd, buf)

    def read(self, count):
        if self.write_fd >= 0:
            try:
                os.close(self.write_fd)
            except:
                pass
            self.write_fd = -1
        return os.read(self.read_fd, count)

    def __del__(self):
        if self.read_fd >= 0:
            try:
                os.close(self.read_fd)
            except:
                pass
        if self.write_fd >= 0:
            try:
                os.close(self.write_fd)
            except:
                pass


class PidFileError(Exception):
    pass


class PidFile:
    """
    Create and maintain a "pid-file" - a file containing the process ID of the
    current instance of a server, acting as a mutex-lock ensuring only one
    server runs at a time, and allowing other processes to signal the daemon.

    This is Unix-specific.
    """

    def __init__(self, pidfilename):
        self._pid_filename = pidfilename

    def getpid(self):
        try:
            pid = open(self._pid_filename,'r').readline()
        except IOError, (eno, estr):
            if eno == errno.ENOENT:
                return 0
            raise PidFileError('%s: %s' % (self._pid_filename, estr))
        try:
            pid = int(pid)
        except ValueError:
            raise PidFileError('%s: unknown format' % \
                               self._pid_filename)
        if pid <= 1:
            raise PidFileError('%s: unknown format' % \
                               self._pid_filename)
        try:
            os.kill(pid,0)
        except OSError, (eno, estr):
            if eno == errno.ESRCH:
                # Stale lock file, remove and sleep to avoid creation race
                print >> sys.stderr, 'Removing state lock for pid %d' % pid
                os.unlink(self._pid_filename)
                time.sleep(2)
                return 0
            raise PidFileError('pid %d: %s' % (pid, estr))

        return pid

    def is_running(self):
        return (self.getpid() != 0)

    def kill(self, sig = signal.SIGTERM):
        pid = self.getpid()
        if pid:
            os.kill(pid, sig)

    def start(self, pid = None):
        if pid is None:
            pid = os.getpid()

        daemon_pid = self.getpid()
        if daemon_pid:
            raise PidFileError('daemon already running, pid %d' % daemon_pid)

        # We use fd ops because we want O_EXCL semantics on the created file
        try:
            pid_fd = os.open(self._pid_filename,
                             os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0644)
        except OSError, (eno, estr):
            raise PidFileError('can\'t create pid file %s: %s' %\
                               (self._pid_filename, estr))
        os.write(pid_fd, '%d\n' % pid)
        os.close(pid_fd)

    def stop(self):
        pid = self.getpid()
        if pid != os.getpid():
            raise PidFileError('Pid file doesn\'t belong to us - pid %d' % pid)
        try:
            os.unlink(self._pid_filename)
        except:
            pass


class Daemon:
    """
    Start a daemon - background, write pidfile, divorce from controlling tty

    This is Unix-specific.
    """

    def __init__(self, worker, pidfile):
        self.worker = worker
        self.pidfile = PidFile(pidfile)

    def cmd_start(self):

        # We want the parent to be able to return useful exit codes to
        # whoever called it, so we use a pipe to pass status from the
        # child back to the parent.
        pipe = Pipe()
        pid = os.fork()
        if pid:
            # Parent
            status = pipe.read(3)
            if status == str(EXIT_OK):
                print 'server pid %d' % pid
                return EXIT_OK
            else:
                return EXIT_ERROR

        # Child
        del pid

        try:
            self.pidfile.start()
        except PidFileError, msg:
            print >> sys.stderr, 'pidfile: %s' % msg
            sys.exit(EXIT_ERROR)

        self.worker.init()

        # Divorce us from the parents controlling TTY (so we don't
        # receive their signals, etc). It's somewhat rude to change
        # these file descriptors from under the python libraries, but
        # it should be safe enough.
        #
        # After this point, anything written to stdout or stderr is lost.
        if 1:
            devnull = os.open('/dev/null', os.O_RDWR)
            os.dup2(devnull, 0)
            os.dup2(devnull, 1)
            os.dup2(devnull, 2)
            os.close(devnull)
            try:
                os.setsid()
            except:
                pass

        signal.signal(signal.SIGTERM, self.sig_stop)

        pipe.write(str(EXIT_OK))            # Signal parent that all is ok
        del pipe

        try:
            self.worker.run()
        except Exception:
            import logging
            logging.exception('Unexpected exception')

        try:
            self.pidfile.stop()
        except PidFileError, msg:
            print >> sys.stderr, 'pidfile: %s' % msg

        sys.exit(EXIT_OK)

    def sig_stop(self, sig, stack):
        self.worker.stop()

    def cmd_stop(self):
        tries = 5
        try:
            pid = self.pidfile.getpid()
        except PidFileError, msg:
            print >> sys.stderr, 'pidfile: %s' % msg
            return EXIT_ERROR
        if pid:
            print 'stopping %d' % pid,
            while self.pidfile.is_running() and tries:
                try:
                    self.pidfile.kill()
                except PidFileError, msg:
                    print >> sys.stderr, 'pidfile: %s' % msg
                    return EXIT_ERROR
                print '.',
                sys.stdout.flush()
                tries = tries - 1
                time.sleep(1)
            if tries:
                print 'done'
                return EXIT_OK
            else:
                print 'failed to kill %s' % pid
        else:
            print 'daemon not running'
        return EXIT_ERROR

    def cmd_status(self):
        pid =  self.pidfile.getpid()
        if pid:
            print 'running, pid is %d' % pid
            return EXIT_OK
        else:
            print 'not running'
            return EXIT_ERROR

    def command(self, cmd_name):
        return getattr(self, 'cmd_' + cmd_name)()
