#
# Copyright 2011 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# Application support which all application types inherit from

from albatross.common import init_logging
from albatross.prepackaged import get_app_classes
from ConfigParser import RawConfigParser, NoSectionError, NoOptionError


class ConfigSection:

    def __init__(self, cfg, section):
        self.cfg = cfg
        self.section = section

    def get_str(self, option, default=None):
        try:
            return self.cfg.get(self.section, option)
        except (NoSectionError, NoOptionError):
            return default

    def get_int(self, option, default=None):
        try:
            return self.cfg.getint(self.section, option)
        except (NoSectionError, NoOptionError):
            return default

    def get_float(self, option, default=None):
        try:
            return self.cfg.getfloat(self.section, option)
        except (NoSectionError, NoOptionError):
            return default

    def get_bool(self, option, default=None):
        try:
            return self.cfg.getboolean(self.section, option)
        except (NoSectionError, NoOptionError):
            return default



class AppConfig(RawConfigParser):

    def __init__(self, fn):
        RawConfigParser.__init__(self)
        self.read([fn])
        # general params
        app_sec = ConfigSection(self, 'application')
        self.page_model = app_sec.get_str('page_model', 'simple')
        self.context_model = app_sec.get_str('context_model', 'hidden-field')
        self.name = app_sec.get_str('name', 'template')
        self.debug = app_sec.get_bool('debug', False)
        self.page_path = app_sec.get_str('page_path', 'pages')
        self.template_path = app_sec.get_str('template_path', 'pages')
        self.module_path = app_sec.get_str('module_path', 'pages')
        self.static_path = app_sec.get_str('static_path', '%s-' % self.name)
        self.start_page = app_sec.get_str('start_page', 'main')
        self.secret = app_sec.get_str('secret', '-=-%s-=-' % self.name)
        # session file params
        self.session_path = app_sec.get_str('session_path', 'session')
        # session server params
        self.session_server = app_sec.get_str('session_server', 'localhost')
        self.server_port = app_sec.get_int('server_port', 34343)
        self.session_age = app_sec.get_int('session_age', 1800)
        # logging params
        log_sec = ConfigSection(self, 'logging')
        self.log_filename = log_sec.get_str('filename')
        self.log_level = log_sec.get_str('level')
        self.log_format = log_sec.get_str('format')
        self.log_datefmt = log_sec.get_str('datefmt')

    def get_app_classes(self):
        return get_app_classes(page_model=self.page_model,
                               context_model=self.context_model)

    def get_app_args(self):
        app_args = {}
        app_args['base_url'] = self.name
        app_args['start_page'] = self.start_page
        app_args['secret'] = self.secret
        # Page model specific args
        if self.page_model == 'simple':
            app_args['template_path'] = self.template_path
        elif self.page_model == 'modular':
            app_args['module_path'] = self.module_path
            app_args['template_path'] = self.template_path
        elif self.page_model == 'random':
            app_args['page_path'] = self.page_path
        # Context model specific args
        if self.context_model == 'hidden-field':
            pass
        elif self.context_model == 'session-file':
            app_args['session_dir'] = self.session_path
            app_args['session_appid'] = self.name
        elif self.context_model in ('session-server', 'branching-server'):
            app_args['session_server'] = self.session_server
            app_args['server_port'] = self.server_port
            app_args['session_appid'] = self.name
        return app_args

    def get_logger(self):
        return init_logging(name=self.name, filename=self.log_filename,
                            format=self.log_format, datefmt=self.log_datefmt)
