#
# Copyright 2009 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# $Id: form.py 16928 2011-04-15 00:13:35Z gregh $

import sys
import string
try:
    set
except NameError:
    from sets import Set as set

from albatross import ApplicationError
from albatross.template import EmptyTag
from albatross.tags import escape


INDENT, POST_INDENT, DEDENT, POST_DEDENT  = range(4)


class CSSStyles:

    def get_style(self, name):
        return self.styles[name]

    def set_style(self, name, style):
        self.styles[name] = style


class DefaultCSSStyles(CSSStyles):

    """
    A collection of named CSS styles used in the HTML output of
    :class:`HTMLTag` instances.

    Each :class:`Form` has a reference to a single instances of this
    class - :attr:`field.styles`.

    Additional styles can be added to the default instance after the
    :class:`Form` is created or an existing instance can be specified
    when the :class:`Form` is created using the *styles* argument.

    The default style names are:

    * buttons
    * legend
    * label
    * field
    * field-error
    * field-invalid,
    * value
    * value-static

    """

    styles = {
        'buttons': 'buttons',
        'legend': 'legend',
        'label': 'label',
        'field': 'field',
        'field-error': 'field-error',
        'field-invalid': 'field-invalid',
        'value': 'value',
        'value-static': 'value-static'
    }


class FormError(Exception):
    """
    Base class for all exceptions raised by the :mod:`albatross.ext.form`
    module.
    """


class FieldValidationError(FormError):

    """
    Raised by :class:`Field` subclasses if :meth:`validate` fails.
    """


class FormValidationError(FormError):

    """
    Raised by :class:`Form` subclasses validation fails for one or
    more fields.

    * *errors* is a dictionary of :exc:`FieldValidationError`
      instances keyed by :class:`Field` instance.

    """

    def __init__(self, errors):
        self.errors = errors
        FormError.__init__(self, 'invalid value for one or more fields')


# Field edit modes.
FORM_CREATE, FORM_UPDATE = range(2)


class HTMLTag:

    """
    Base class for any class which provides HTML output. Arbitrary
    HTML attributes can be specified as *html_attrs*.
    """

    def __init__(self, html_attrs=None):
        self.html_attrs = html_attrs

    def format_tag(self, tag, closed=False, **attrs):
        if self.html_attrs:
            attrs.update(self.html_attrs)
        parts = [tag]
        for attr, value in attrs.iteritems():
            if attr == 'cssclass':
                attr = 'class'          # side-step "class" reserved word
            if value is None:
                parts.append(str(attr))
            elif isinstance(value, bool):
                if value:
                    parts.append(str(attr))
            else:
                parts.append('%s="%s"' % (attr, escape(value)))
        if closed:
            return '<%s />' % (' '.join(parts))
        else:
            return '<%s>' % (' '.join(parts))

    def to_html(self, ctx, form):
        raise NotImplementedError('%s: to_html method must be implemented by '
                                  'subclass' % self.__class__)

    def set_disabled(self, is_disabled):
        pass


class DataHTMLTag(HTMLTag):
    """
    Output HTML for a value which which can be loaded from another object.

    * *attr* is the name of the attribute in the model
      manipulated by :meth:`load` and :meth:`merge`.

    """
    def __init__(self, attr=None, value=None, html_attrs=None):
        HTMLTag.__init__(self, html_attrs=html_attrs)
        self.attr = attr
        self.value = value
        self.is_display_value_set = False

    def load(self, load_obj):
        if self.attr is None:
            return
        try:
            self.set_value(getattr(load_obj, self.attr))
        except (AttributeError, TypeError), e:
            msg = '%s [%s.load(): load_obj %s, attr %r]' %\
                (e, self.__class__, load_obj.__class__, self.attr)
            raise e.__class__, msg, sys.exc_info()[2]

    def merge(self, merge_obj):
        if self.attr is None:
            return
        try:
            setattr(merge_obj, self.attr, self.get_value())
        except (AttributeError, TypeError, ValueError), e:
            msg = '%s [%s.merge(): merge_obj %s, attr %r, value %r]' %\
                (e, self.__class__, merge_obj.__class__, self.attr, self.value)
            raise e.__class__, msg, sys.exc_info()[2]

    def get_value(self):
        "Return the converted value in this field"
        if self.is_display_value_set:
            return self.get_merge_value(self.value)
        else:
            return self.value

    def set_value(self, value, is_display_value=False):
        self.value = value
        self.is_display_value_set = is_display_value

    def get_merge_value(self, s):
        """
        This should return the correctly typed object to update the field in
        the original object.  Override this if you are using a non-string
        value.
        """
        return s

    def get_display_value(self, ctx, form):
        """
        Override this method to provide any required formatting or modification
        of the displayed value.
        """
        # Use single nonblocking space if value is empty or whitespace to
        # prevent table cell collapsing. This is really only needed for IE,
        # which does not support CSS2's "empty-cells: show".
        if self.value is not None:
            value = str(self.value).strip()
            if value:
                return escape(value)
        return '&nbsp;'


class HTMLTreeTag(HTMLTag):

    """
    An :class:`HTMLTag` with children.
    """

    def __init__(self, children, html_attrs=None):
        HTMLTag.__init__(self, html_attrs=html_attrs)
        self.children = children

    def load(self, load_obj):
        for child in self.children:
            child.load(load_obj)

    def merge(self, merge_obj):
        for child in self.children:
            child.merge(merge_obj)

    def register_fields(self, form):
        for child in self.children:
            child.register_fields(form)

    def to_html(self, ctx, form):
        for child in self.children:
            child.to_html(ctx, form)

    def set_disabled(self, is_disabled):
        for child in self.children:
            child.set_disabled(is_disabled)


class Table(HTMLTreeTag):

    """
    HTML `TABLE` tag.
    """

    def __init__(self, header_row, children, html_attrs=None):
        HTMLTreeTag.__init__(self, children, html_attrs=html_attrs)
        self.header_row = header_row

    def to_html(self, ctx, form):
        form.write_content(ctx, self.format_tag('table'), POST_INDENT)
        if self.header_row:
            self.header_row.to_html(ctx, form)
        for child in self.children:
            child.to_html(ctx, form)
        form.write_content(ctx, '</table>', DEDENT)

    def append(self, child):
        self.children.append(child)

    def remove(self, child):
        self.children.remove(child)


class Row(HTMLTreeTag):

    """
    HTML `TR` tag.
    """

    def to_html(self, ctx, form):
        tag = self.format_tag('tr', valign="top")
        form.write_content(ctx, tag, POST_INDENT)
        for child in self.children:
            child.to_html(ctx, form)
        form.write_content(ctx, '</tr>', DEDENT)


class Col(HTMLTreeTag):

    """
    HTML `TD` tag.
    """

    def to_html(self, ctx, form):
        form.write_content(ctx, self.format_tag('td'), POST_INDENT)
        for child in self.children:
            child.to_html(ctx, form)
        form.write_content(ctx, '</td>', DEDENT)


class HeaderCol(HTMLTreeTag):

    """
    HTML `TH` tag.
    """

    def to_html(self, ctx, form):
        form.write_content(ctx, self.format_tag('th'), POST_INDENT)
        for child in self.children:
            child.to_html(ctx, form)
        form.write_content(ctx, '</th>', DEDENT)


class Label(HTMLTag):

    """
    Simple displayed string value with no label.
    """

    def __init__(self, value, html_attrs=None):
        HTMLTag.__init__(self, html_attrs=html_attrs)
        self.value = value

    def get_display_value(self, ctx, form):
        if self.value.strip():
            return escape(self.value)
        return '&nbsp;'

    def to_html(self, ctx, form):
        form.write_content(ctx, self.get_display_value(ctx, form))

    def register_fields(self, form):
        pass

    def load(self, load_obj):
        pass


class Form(HTMLTreeTag):

    """
    Encapsulates an HTML form and its input fields.

    * *legend* is displayed as the title of the form.

    * *children* is a list of :class:`HTMLTag` subclasses including
      all the :class:`Form` fields and markup fields for layout.

    * *buttons* is an optional :class:`Buttons` instance.

    * *styles* is an optional :class:`CSSStyles` instance. If not
      specified a :class:`DefaultCSSStyles` instance is created.

    """

    def __init__(self, legend, children,
                 buttons=None, styles=None, html_attrs=None):
        HTMLTreeTag.__init__(self, children, html_attrs)
        self.legend = legend
        self.children = children
        self.buttons = buttons
        self.styles = styles and styles or DefaultCSSStyles()
        self.disabled = False
        self.validation_errors = {}
        self.valid = True
        self.input_name = None
        self.show_errors = False
        self.static = False
        self.fields = set()
        self.edit_mode = FORM_CREATE
        self.register_fields(self)

    def register_field(self, name, field):
        if not field in self.fields:
            self.fields.add(field)
            setattr(self, name, field)

    def get_field(self, name):
        for field in self.fields:
            if field.name == name:
                return field

    def clear(self):
        """
        Clear all field values, set form and all fields to valid
        state, clear disabled and set edit mode to
        :const:`FORM_CREATE`.
        """
        self.clear_errors()
        for field in self.fields:
            if not field.static:
                field.clear()
        self.disabled = False
        self.edit_mode = FORM_CREATE

    def clear_errors(self):
        """
        Clear errors and set all fields to valid.
        """
        for field in self.fields:
            field.valid = True
        self.validation_errors = {}
        self.valid = True

    def merge(self, merge_obj=None):
        """
        Merge all field values to *merge_obj*.
        """
        for field in self.fields:
            if not field.static:
                field.merge(merge_obj)

    def load(self, load_obj):
        """
        Load field values from *load_obj*.
        """
        for field in self.fields:
            field.load(load_obj)
        self.edit_mode = FORM_UPDATE

    def validate(self):
        """
        Call :meth:`validate` on each field. Sets *self.valid* to
        False and raises :exc:`FormValidationError` if any fields are
        invalid. Validation exceptions are stored in
        *self.validation_errors* and are cleared by calling
        :meth:`clear`, :meth:`clear_errors` or on the next call to
        :meth:`validate`.

        Static fields or fields of a static form are not validated.
        """
        if self.static:
            return
        self.valid = True
        self.validation_errors = {}
        for f in self.fields:
            if f.static:
                continue
            if f.value is None:
                f.value = ''
            try:
                f.validate(self, f.value)
                f.valid = True
            except FieldValidationError, e:
                f.valid = False
                # when a subform reraises a validation error, it
                # sets the message to None so that the higher level
                # forms see that there's an error but don't act on it
                if e is not None:
                    self.validation_errors[f] = e
        if self.validation_errors:
            self.valid = False
            raise FormValidationError(self.validation_errors)

    def get_validation_error(self, field):
        return self.validation_errors[field]

    def set_disabled(self, is_disabled):
        "Disable or enable data entry on all fields on this form."
        self.disabled = is_disabled
        HTMLTreeTag.set_disabled(self, is_disabled)

    def to_html(self, ctx):
        """
        Render all fields and buttons as HTML. This is an internal
        method. To render the form use :meth:`run`.
        """
        for child in self.children:
            child.to_html(ctx, self)
        if self.buttons:
            self.buttons.to_html(ctx, self)

    def indent(self):
        self._indent += 1

    def dedent(self):
        self._indent -= 1
        if self._indent < 0:
            self._indent = 0

    def write_content(self, ctx, content, indent=None):
        """
        A wrapper for :meth:`ctx.write_content` which provides
        indentation to assist in generating human readable HTML
        output.

        Valid values for *indent* are:

        * INDENT
        * DEDENT
        * POST_INDENT
        * POST_DEDENT

        All other values are ignored.
        """
        if indent == INDENT:
            self.indent()
        elif indent == DEDENT:
            self.dedent()
        ctx.write_content(' ' * (self._indent * 2))
        ctx.write_content('%s\n' % content)
        if indent == POST_INDENT:
            self.indent()
        elif indent == POST_DEDENT:
            self.dedent()

    def run(self, ctx, name, opts):
        """
        Called by alx-form tag handler to render the form. This method
        sets the correct internal state for rendering, resets
        indentation and registers input fields before calling :meth:`to_html`.
        """
        self.input_name = name
        self.opts = opts
        self._indent = 0
        self.to_html(ctx)


class FormModel:

    """
    Convenience class for simple form/model manipulation.
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class LabelFieldRow(Row):

    """
    Build a two column Row instance with a name Label and a Field.
    """

    def __init__(self, field, label=None):
        if label is None:
            label = field.name
        Row.__init__(self,
                     (Col((Label(label),)),
                      Col((field,))))


class FieldsetForm(Form):

    """
    A :class:`Form` subclass which wraps the HTML output in a
    `<FIELDSET>` tag.
    """

    def to_html(self, ctx):
        # This all looks a bit crufty but it's because of fieldset rendering
        # problems in IE 5/6/7/8.
        self.write_content(ctx, self.format_tag('fieldset'), POST_INDENT)
        self.write_content(ctx, '<span class="%s">%s</span>' %
                           (escape(self.styles.get_style('legend')),
                            escape(self.legend)))
        Form.to_html(self, ctx)
        self.write_content(ctx, '</fieldset>', DEDENT)


class Fieldset(HTMLTreeTag):

    """
    Container for a collection of :class:`Field` instances.

    Writes a table of fields. Note that the <fieldset> HTML tags are
    intentionally provided by the :class:`FieldsetForm` and not this
    class.
    """

    def to_html(self, ctx, form):
        form.write_content(ctx, self.format_tag('table'), POST_INDENT)
        for child in self.children:
            child.to_html(ctx, form)
        form.write_content(ctx, '</table>', DEDENT)


class Field(DataHTMLTag):

    """
    Base class for input fields.

    * *name* is the display name tyically shown to the left of the
      input control in the generated HTML.

    * Specify *merge_obj* if this :class:`Field` is loaded/merged from
      a model object other than the one used for the :class:`form` (this
      should usually be left as `None`).

    * Set *required* to `True` if a value must be entered for this
      :class:`Field` (a :exc:`FieldValidationError` will be raised on a
      call to :meth:`validate` if the :class:`Field` is empty).

    * If *static* is `True` the :class:`Field` is always displayed in
      static report style preventing user input.

    Creating your own Field-subclasses

    Several methods may be provided by custom subclasses as needed:

    * *get_display_value(self, ctx, form)*
        should return the value stored in *self.value* to how you
        want it displayed in the browser.  This is typically called
        during to_html and the value of *self.value* will be replaced
        with this after that point.

    * *get_merge_value(self, s)*
        Should convert the string version of the value passed in and
        return a value of the appropriate type.

    * *validate(self, form)*
        Validate the string version of the value stored in the field.
        The method should just return if it validates correctly; if not,
        raise :exc:`FormValidationError` with an appropriate error
        message.

        :class:`Field`'s that are static or are members of a static form
        are not validated.

    * *to_html(self, ctx, form)*
        Generate custom HTML for this tag.  Use:

            form.write_content(ctx, '<your html here>')

        to write your HTML to the output stream.
    """

    type = None

    def __init__(self, name, attr,
                 merge_obj=None, value=None, html_attrs=None,
                 required=False, static=False):
        self.static = static
        DataHTMLTag.__init__(self, attr, value=value, html_attrs=html_attrs)
        self.name = name
        self.merge_obj = merge_obj
        self.valid = True
        self.required = required

    def get_merge_value(self, value):
        """
        Override :meth:`get_merge_value` where the value received
        from Albatross on a form submission should be checked or
        modified before being stored.
        """
        return value

    def register_fields(self, form):
        # note that fields are registered whether they're static or not
        # because we might need to reference them via get_field and so
        # that they're initialised correctly during load()
        form.register_field(self.input_name(), self)

    def validate(self, form, s):
        """
        Subclasses should use this method to validate *s*
        after form submission.  The method should raise a
        :exc:`FieldValidationError` for invalid values.
        """
        return

    def load(self, form_load_obj):
        """
        Load *self.value* from a named attribute in the model.
        """
        if self.merge_obj is not None:
            form_load_obj = self.merge_obj
        DataHTMLTag.load(self, form_load_obj)

    def merge(self, form_merge_obj=None):
        """
        Merge value back into the model.
        """
        if self.merge_obj is not None:
            form_merge_obj = self.merge_obj
        DataHTMLTag.merge(self, form_merge_obj)

    def clear(self):
        self.set_value(None)

    _input_name_tx_table = string.maketrans(' ./,', '_' * 4)
    def input_name(self):
        # XXX this could be improved ...
        return self.name.lower().translate(self._input_name_tx_table)

    def input_name_for_form(self, form):
        """
        Create a unique string representation of the field to be used
        as an Albatross input name.
        """
        return "%s.%s.value" % (form.input_name, self.input_name())

    def write_static_html(self, ctx, form):
        """
        Write a static HTML representation of the field. This method
        is called by :meth:`to_html` if *self.static* is True of if
        the whole :class:`Form` is being statically rendered.
        """
        form.write_content(ctx, self.get_display_value(ctx, form))

    def write_form_html(self, ctx, form):
        """
        Write the interactive form control HTML representation of the
        field.
        """
        raise NotImplementedError('%s: method must be implemented by '
                                  'subclass' % self.__class__)

    def write_errors_html(self, ctx, form):
        """
        Write the HTML representation of the
        :exc:`FieldValidationError` for this :class:`Field`. Called
        during HTML generation if this :class:`Field` raised an
        exception during :class:`Form` validation.
        """
        form.write_content(ctx, '<tr valign="top">', POST_INDENT)
        form.write_content(ctx,
            '<td></td><td><span class="%s">%s</span></td>' %
            (form.styles.get_style('field-error'),
             escape(form.get_validation_error(self))))
        form.write_content(ctx, '</tr>', DEDENT)

    def to_html(self, ctx, form):
        """
        Field subclasses must provide different output for static,
        enabled and disabled states and may output validation error
        messages.

        The default implementation of :meth:`write_static_html` and
        :meth:`write_errors_html` should suffice for most cases. A
        typical subclass will override :meth:`write_form_html` to
        provide the HTML ouput for the input controls required (text
        input, checkbox, etc.)
        """
        if self.static or form.opts.static or form.disabled:
            self.write_static_html(ctx, form)
        else:
            self.write_form_html(ctx, form)
            if form.opts.show_errors and not self.valid:
                self.write_errors_html(ctx, form)


class InputField(Field):

    """
    Base class for generating an HTML <input> tag.  Requires that the
    instance has a *self.type* which is the type of the HTML input, ie,
    it will generate <input type="(*self.type*)" ...>.

    Subclasses are expected to maintain *self.value* as the correct type
    (eg, *string* (for text), *int*, *datetime*, etc) and to do validation
    of the input as required.  *Self.value* can also contain invalid
    strings if the user has partly edited a field.  Calling *merge* or
    *get_value* before *validate* has succeeded may raise an exception.

    The conversion of the internal value to the display value is complicated
    because Albatross sets the value of an empty field to None so that needs
    to be handled or all empty text fields in a form are converted to "None".
    """

    def write_form_html(self, ctx, form):
        input_name = self.input_name_for_form(form)
        if not form.disabled:
            ctx.input_add(self.type, input_name, self.value)
        if not self.is_display_value_set or self.value is None:
            self.set_value(self.get_display_value(ctx, form),
                           is_display_value=True)
        tag = self.format_tag('input', closed=True, type=self.type,
                              name=input_name, value=self.value,
                              disabled=bool(form.disabled))
        form.write_content(ctx, tag)


class TextField(InputField):

    """
    HTML <input type="text"> tag. Merged data is a string. If
    *self.required* is `True` :meth:`validate` will raise a
    :exc:`FieldValidationError` if the input field is empty. No other
    validation or conversion is performed.
    """

    type = 'text'

    def get_merge_value(self, value):
        """
        Empty input field is submitted as None - convert to empty
        string.
        """
        if value is None:
            # overwrite the returned None or it will reappear
            # as "None" next time the form is displayed
            self.set_value('')
            return ''
        return str(value)

    def load(self, load_obj):
        InputField.load(self, load_obj)
        if self.value is None:
            value = ''
        else:
            value = str(self.value)
        self.set_value(value)

    def validate(self, form, s):
        if self.required and s.strip() == '':
            raise FieldValidationError('A value is required for "%s"' % self.name)

    def get_display_value(self, ctx, form):
        if self.static or form.static:
            if self.value is None:
                return '&nbsp;'
            return InputField.get_display_value(self, ctx, form)
        else:
            if self.value is None:
                return ''
            return self.value


class PasswordField(TextField):

    """
    A :class:`TextField` with text input hidden by '*'
    """

    type = 'password'


class StaticField(Field):

    """
    String :class:`Field` which never accepts user input.
    """

    def __init__(self, name, value, html_attrs=None):
        Field.__init__(self, name, None, value=value,
                       html_attrs=html_attrs, static=True)

    def to_html(self, ctx, form):
        self.write_static_html(ctx, form)

    def load(self, load_obj):
        pass


class Textarea(TextField):

    """
    An HTML `TEXTAREA` tag. Validate and merge rules are the same as
    :class:`TextField`.
    """

    def get_display_value(self, ctx, form):
        value = TextField.get_display_value(self, ctx, form)
        # add line break b/w lines so static appearance resembles
        # textarea input control.
        return '<br />'.join(value.splitlines())

    def write_form_html(self, ctx, form):
        input_name = self.input_name_for_form(form)
        if not form.disabled:
            ctx.input_add(self.type, input_name, self.value)
        if not self.is_display_value_set:
            self.set_value(self.get_display_value(ctx, form),
                           is_display_value=True)
        tag = self.format_tag('textarea', name=input_name,
                              disabled=bool(form.disabled))
        form.write_content(ctx, '%s%s</textarea>' % (tag, escape(self.value)))


class IntegerField(InputField):

    """
    A :class:`InputField` which validates and merges integer values.
    """

    type = 'text'

    def validate(self, form, s):
        try:
            value = s.strip()
            if not self.required and value == '':
                return
        except AttributeError:
            # value is not a string
            value = s
        try:
            int(value)
        except ValueError:
            raise FieldValidationError('An integer value is required '
                                       'for "%s"' % self.name)

    def get_merge_value(self, s):
        s = s.strip()
        if not self.required and s == '':
            return
        return int(s)

    def get_display_value(self, ctx, form):
        if self.value is None:
            return '0'
        return '%s' % self.value


class FloatField(InputField):

    """
    A :class:`TextField` which validates and merges float values.
    """

    type = 'text'

    def validate(self, form, s):
        try:
            value = s.strip()
            if not self.required and value == '':
                return
        except AttributeError:
            # value is not a string
            value = s
        try:
            f = float(value)
        except ValueError:
            raise FieldValidationError('A float value is required '
                                       'for "%s"' % self.name)

    def get_merge_value(self, s):
        s = s.strip()
        if not self.required and s == '':
            return
        return float(s)

    def get_display_value(self, ctx, form):
        if self.value is None:
            return '0.0'
        return '%s' % self.value


class Checkbox(InputField):

    """
    An HTML <input type="checkbox"> control. *self.value* is always
    valid and the merged value is boolean.
    """

    type = 'checkbox'
    checkbox_on = 'on'

    def validate(self, form, s):
        pass

    def get_display_value(self, ctx, form):
        # in an <input type="checkbox"> field, we only care if the value
        # field is set to "on"; anything else we consider off.  When
        # checked, the value field is put into the name attribute; otherwise
        # it's empty.
        if self.value:
            return self.checkbox_on
        return 'off'

    def get_merge_value(self, s):
        return (s == self.checkbox_on)

    def write_form_html(self, ctx, form):
        input_name = self.input_name_for_form(form)
        if not form.disabled:
            ctx.input_add(self.type, input_name, self.value)
        if not self.is_display_value_set:
            self.set_value(self.get_display_value(ctx, form),
                           is_display_value=True)
        tag = self.format_tag('input', closed=True, type=self.type,
                              name=input_name, value=self.checkbox_on,
                              checked=bool(self.get_value()),
                              disabled=bool(form.disabled))
        form.write_content(ctx, tag)

    def write_static_html(self, ctx, form):
        tag = self.format_tag('input', closed=True, type=self.type,
                              value=self.checkbox_on,
                              checked=bool(self.get_value()),
                              disabled=True)
        form.write_content(ctx, tag)


class OptionsField(Field):

    """
    Base class to handle selection of a single value from a list of
    options, ie, for a select or radio list.
    """

    def __init__(self, name, options, attr,
                 merge_obj=None, value=0, html_attrs=None, static=False):
        self.options = options
        Field.__init__(self, name, attr,
                       merge_obj=merge_obj, value=value,
                       html_attrs=html_attrs, static=static)

    def set_options(self, options):
        "change the *options* displayed by the select field."
        self.options = options
        self.value = 0

    def get_merge_value(self, value):
        # XXX does this work if (index, option) is (str, str)?
        if not self.options:
            return None
        try:
            value = int(value)
        except (ValueError, TypeError):
            value = 0
        for index, option in self.options:
            if value == index:
                return index
        index, option = self.options[0]
        return index

    def get_display_value(self, ctx, form):
        display_value = ''
        if self.options:
            value = str(self.value)
            for index, option in self.options:
                if value == str(index):
                    display_value = option
                    break
        return escape(display_value)


class SelectField(OptionsField):

    """
    Manage a drop down list of options.

    * *options* is a list of tuples of (value, display value)
    """

    def write_form_html(self, ctx, form):
        if not self.options:
            return self.write_static_html(ctx, form)
        input_name = self.input_name_for_form(form)
        ctx.input_add(self.type, input_name, self.value)
        tag = self.format_tag('select', name=input_name,
                              disabled=bool(form.disabled))
        form.write_content(ctx, tag, POST_INDENT)
        value = str(self.get_value())
        for index, option in self.options:
            if value == str(index):
                tag = self.format_tag('option', value=index, selected=None)
            else:
                tag = self.format_tag('option', value=index)
            option_tag_parts = (
                tag,
                escape(option),
                self.format_tag('/option')
            )
            form.write_content(ctx, ''.join(option_tag_parts))
        self.set_value(str(self.value), is_display_value=True)
        form.write_content(ctx, '</select>', DEDENT)


class RadioField(OptionsField):

    """
    Manage a radio list.

    * *options* is a list of tuples of (value, display value)

    When the field or the enclosing form is static, we just emit the
    selected option.  (XXX I hope that's the right thing to do)
    """

    type = 'radio'

    def write_form_html(self, ctx, form):
        if not self.options:
            return self.write_static_html(ctx, form)
        input_name = self.input_name_for_form(form)
        ctx.input_add(self.type, input_name, self.value)
        value = str(self.get_value())
        for index, option in self.options:
            tag = self.format_tag('input', type=self.type,
                                  name=input_name, value=index,
                                  checked=(value == str(index)),
                                  disabled=bool(form.disabled))
            radio_tag_parts = (
                tag,
                escape(option),
                self.format_tag('br')
            )
            form.write_content(ctx, ''.join(radio_tag_parts))
        self.set_value(str(self.value), is_display_value=True)


class Buttons(HTMLTag):

    """
    A collection of :class:`Button` instances.
    """

    def __init__(self, buttons, html_attrs=None):
        self.buttons = buttons
        HTMLTag.__init__(self, html_attrs)

    def register_fields(self, form):
        pass

    def to_html(self, ctx, form):
        if form.opts.static:
            return
        tag = self.format_tag('div', cssclass=form.styles.get_style('buttons'))
        form.write_content(ctx, tag, POST_INDENT)
        for button in self.buttons:
            ctx.input_add('submit', button.name, button.value)
            button.to_html(ctx, form)
            form.write_content(ctx, '&nbsp;')
        form.write_content(ctx, '</div>', DEDENT)


class Button(HTMLTag):

    """
    An HTML `submit` button. All buttons in a :class:`Form` should be
    collected in a single :class:`Buttons` instance which is passed to
    the constructor of the :class:`Form`.

    * *value* is written on the face of the button when displayed.

    * *name* is returned from the browser and can be tested in the
      application using *ctx.req_equals('name')*.
    """

    def __init__(self, value, name, html_attrs=None):
        self.value = value
        self.name = name
        HTMLTag.__init__(self, html_attrs)

    def register_fields(self, form):
        pass

    def to_html(self, ctx, form):
        if form.opts.static:
            return
        tag = self.format_tag('input', closed=True, type="submit",
                              name=self.name, value=self.value,
                              disabled=bool(form.disabled))
        form.write_content(ctx, tag)


class PagedListBase:

    """
    An IteratorTable item list which supports pagination. Subclasses
    should implement page_count() and items_on_page() to create
    @row_class instances of the items for a given page on demand.
    """

    def __init__(self, pagesize=None):
        self.pagesize = pagesize

    def get_column_names(self):
        raise NotImplementedError(
            '%s: get_column_names method must be implemented by '
            'subclass' % self.__class__)

    def items_on_page(self, page):
        raise NotImplementedError(
            '%s: items_on_page method must be implemented by '
            'subclass' % self.__class__)

    def page_count(self):
        raise NotImplementedError(
            '%s: paged_count method must be implemented by '
            'subclass' % self.__class__)

    def num_children(self):
        raise NotImplementedError(
            '%s: num_children method must be implemented by'
            'subclass' % self.__class__)

    def get_child(self, index):
        raise NotImplementedError(
            '%s: get_child method must be implemented by'
            'subclass' % self.__class__)

    def get_children(self):
        raise NotImplementedError(
            '%s: get_children method must be implemented by'
            'subclass' % self.__class__)

    def append(self, item):
        raise NotImplementedError(
            '%s: append method must be implemented by '
            'subclass' % self.__class__)

    def remove(self, item):
        raise NotImplementedError(
            '%s: remove method must be implemented by '
            'subclass' % self.__class__)

    def start_index_of_page(self, page):
        if self.pagesize:
            if page < 1:
                page = 1
            elif page > self.page_count():
                page = self.page_count()
            return self.pagesize * (page - 1)
        else:
            return 0

    def has_next(self, page):
        return (self.pagesize > 0 and page < self.page_count())

    def has_prev(self, page):
        return (self.pagesize > 0 and page > 1)


class PagedList(PagedListBase):

    """
    Simple pager which stores all items and returns the items for a
    given page in a @row_class instance on demand.
    """

    def __init__(self, items, row_class=None, pagesize=None):
        if row_class is None:
            row_class = IteratorTableRow
        PagedListBase.__init__(self, pagesize=pagesize)
        self.row_class = row_class
        self.items = items

    def get_column_names(self):
        return [field.children[0].name for field in self.row_class().children]

    def items_on_page(self, page):
        if self.pagesize > 0:
            start_index = self.start_index_of_page(page)
            end_index = start_index + self.pagesize
            if end_index > len(self.items):
                end_index = len(self.items)
            items = self.items[start_index:end_index]
        else:
            items = self.items
        return [self.row_class(item) for item in items]

    def page_count(self):
        if self.pagesize > 0 and self.items:
            return (len(self.items) - 1) / self.pagesize + 1
        else:
            return 1

    def num_children(self):
        return len(self.items)

    def get_child(self, index):
        return self.items[index]

    def get_children(self):
        return self.items

    def append(self, item):
        self.items.append(item)
        return self.row_class(item)

    def remove(self, item):
        self.items.remove(item)

    def update(self, item, *attrs):
        for i, check_item in enumerate(self.items):
            match = True
            for attr in attrs:
                if not hasattr(check_item, attr) or \
                        getattr(check_item, attr) != getattr(item, attr):
                    match = False
                    break
            if match:
                check_item.__dict__.update(item.__dict__)
                return i
        raise FormError('item %r not found in table' % item)


class PageSelectionDisplayBase:
    """
    Base class for displaying page selection
    """
    def write_page_selection_above(self, ctx, form,
                                   input_name, paged_children, page):
        raise NotImplementedError('%s: method must be implemented'
                                  ' by subclass' % self.__class__)

    def write_page_selection_below(self, ctx, form,
                                   input_name, paged_children, page):
        raise NotImplementedError('%s: method must be implemented'
                                  ' by subclass' % self.__class__)

    def write_link_for_page(self, ctx, form, input_name, page, title):
        form.write_content(ctx, '<a href="?%s=%s">%s</a>' %
                                (input_name, page, title))


class PrevNextPageSelectionDisplay(PageSelectionDisplayBase):

    def write_page_selection_above(self, ctx, form,
                                   input_name, paged_children, page):
        pass

    def write_page_selection_below(self, ctx, form,
                                   input_name, paged_children, page):
        if paged_children.has_prev(page):
            self.write_link_for_page(ctx, form, input_name, page - 1, 'Prev')
        if paged_children.has_next(page):
            self.write_link_for_page(ctx, form, input_name, page + 1, 'Next')


class IteratorTable(Field):
    """
    Create an HTML table from a list of objects.

    * *table_attr* is the name of the instance var in the class which
      creates this object.  This is used when directing Albatross to update the
      form values.

    * *header_row* is a :class:`Row` of :class:`HeaderCol` instances (usually)
      which are used to put a header on the table.  An empty list or None will
      suppress any headers.

    * *row_class* is a subclass of :class:`IteratorTableRow`.  It is used to
      render each row in the table.  It will be used to contruct each row by
      being instantiated with each element of *content_list* in turn.

      If you need pass the row class constructor some extra arguments from the
      table constructor, assign them to instance variables and make the
      *row_class* a bound method to a method that marshalls the arguments
      before calling the row constructor.

    * *pager* is the PagedListBase subclass which provides paginated
       data to display in the table.

    """
    def __init__(self, table_attr, header_row, pager, html_attrs=None,
                 page_selection_display=PrevNextPageSelectionDisplay()):
        # XXX rather than __class__.__name__, should pass in a name
        Field.__init__(self, self.__class__.__name__, table_attr,
                       html_attrs=html_attrs)
        self.table_attr = table_attr
        if not isinstance(pager, PagedListBase):
            raise FormError('pager %r is not a subclass of PagedList' %
                            pager)
        self.pager = pager
        self.page = 1
        self.page_selection_display = page_selection_display
        self.current_page_items = None
        if header_row is None:
            headers = []
            for name in self.pager.get_column_names():
                headers.append(HeaderCol((Label(name),)))
            self.header_row = Row(headers)
        else:
            self.header_row = header_row

    def to_html(self, ctx, form):
        # if there's a pagesize set on the pager use that, otherwise
        # use the form's option
        if not self.pager.pagesize and form.opts.pagesize:
            self.pager.pagesize = form.opts.pagesize
            self.current_page_items = None       # force page reload
        input_name = '%s.%s.page' % (form.input_name, self.table_attr)
        # if there's a pagesize try picking up the page number from
        # the request
        if self.pager.pagesize:
            if ctx.request.has_field(input_name):
                ctx.merge_vars(input_name)       # (this will update
                                                 #  self.page)
            try:
                page = int(self.page)
            except (ValueError, TypeError):
                page = 1
        else:
            page = 1
        if page != self.page or not self.current_page_items:
            self.load_page(page)
        self.set_disabled(form.disabled)
        # emit prev/next to go to page
        if self.pager.pagesize:
            self.page_selection_display.write_page_selection_above(
                ctx, form, input_name, self.pager, self.page)
        # generate the table itself
        form.write_content(ctx, self.format_tag('table'), POST_INDENT)
        if self.header_row:
            self.header_row.to_html(ctx, form)
        for i, child in enumerate(self.current_page_items):
            child.input_name = '%s.%s.current_page_items[%d]' % (
                form.input_name, self.table_attr, i)
            child.run(ctx, child.input_name, form.opts)
        form.write_content(ctx, '</table>', DEDENT)
        # emit prev/next to go to page
        if self.pager.pagesize:
            self.page_selection_display.write_page_selection_below(
                ctx, form, input_name, self.pager, self.page)

    def goto_page(self, page):
        "Jump to specified page.  If page == -1, go to last page."
        if not self.pagesize:
            return
        if page < 0 or page > self.pager.page_count():
            self.page = self.pager.page_count()
        else:
            self.page = page
        self.load_page()

    def load_page(self, page=None):
        if page is None:
            page = self.page
        else:
            self.page = page
        self.current_page_items = self.pager.items_on_page(page)

    def append(self, item):
        "Append a new item of data to the table"
        if self.current_page_items is None:
            self.load_page()
        self.current_page_items.append(self.pager.append(item))

    def remove(self, item):
        "Remove an item of data from the table"
        try:
            self.pager.remove(item)
            self.load_page()
        except ValueError, e:
            raise FormError('item %r not found in table' % item)

    def update(self, item, *attrs):
        "Update an item of data in the table"
        if self.pager.update(item, *attrs):
            self.load_page()

    def load(self, load_obj):
        children = self.pager.items_on_page(self.page)
        assert len(load_obj) == len(children)
        for child, child_load_obj in zip(children, load_obj):
            child.load(child_load_obj)

    def merge(self, merge_obj):
        assert len(merge_obj) == len(self.current_page_items)
        for child, child_merge_obj in zip(children, merge_obj):
            child.merge(child_merge_obj)

    def clear(self):
        if self.current_page_items:
            for child in self.current_page_items:
                child.clear()

    def set_disabled(self, is_disabled):
        if self.current_page_items:
            for child in self.current_page_items:
                child.set_disabled(is_disabled)

    def input_name(self):
        return self.table_attr

    def register_fields(self, form):
        form.register_field(self.table_attr, self)
        if self.current_page_items:
            page_offset = self.pager.start_index_of_page(self.page)
            for i, child in enumerate(self.current_page_items):
                # update the title of the subform and reregister fields
                child.legend = '%s.%s.children[%d]' %\
                    (form.input_name, self.table_attr, page_offset+i)
                child.register_fields(child)

    def validate(self, form, s):
        self.valid = True
        validation_errors = {}
        form.register_field(self.table_attr, self)
        for child in self.current_page_items:
            try:
                child.validate()
            except FormValidationError, e:
                child.valid = False
                validation_errors[child] = e
        if validation_errors:
            self.valid = False
            raise FormValidationError(validation_errors)

    def num_children(self):
        return self.pager.num_children()

    def get_children(self):
        return self.pager.get_children()


class IteratorTableRow(Form):
    """
    When using an IteratorTable, the class that's used to display each row
    must be a subclass of IteratorTableRow so that the rendering and
    validation is performed correctly.  This is checked for in the
    ``IteratorTable`` constructor.

    Each row in a table is actually treated as a separate sub-``Form``.
    This allows Albatross to traverse the table hierarchy when it's
    updating the field values.
    """
    # The gist of what happens is that a new form is created for each iterator
    # row instance.  Each field within the row is added to the iterator row
    # class.  When Albatross goes to update the field values, it finds the
    # correct vars in the array element.
    #
    # That's as clear as mud.
    def __init__(self, cols):
        Form.__init__(self, 'IteratorTableRow', cols)

    def to_html(self, ctx):
        "generates the row's content bracketed by <tr>/</tr>"
        tag = self.format_tag('tr', valign="top")
        self.write_content(ctx, tag, POST_INDENT)
        for cols in self.children:
            cols.to_html(ctx, self)
        self.write_content(ctx, '</tr>', DEDENT)

    def validate(self):
        # when a subform reraises a validation error, it sets the message to
        # None so that the higher level forms see that there's an error but
        # don't act on it (see Form.validate()).
        try:
            Form.validate(self)
        except FormValidationError:
            raise FormValidationError(None)


class FormOptions:

    """
    Collect the attributes we're interested in from the alx-form tag.
    """

    def __init__(self, show_errors=False, static=False, pagesize=None):
        self.show_errors = show_errors
        self.static = static
        if pagesize:
            try:
                self.pagesize = int(pagesize)
            except ValueError, e:
                raise ApplicationError('%s: "%s"' % (e, pagesize))
        else:
            self.pagesize = None


class FormTag(EmptyTag):

    name = 'alx-form'

    def to_html(self, ctx):
        if not self.has_attrib('name'):
            raise ApplicationError('alx-form: no "name" attribute')
        form = self.eval_attrib(ctx, 'name')
        opts = FormOptions(show_errors=self.get_bool_attrib(ctx, 'errors'),
                           static=self.get_bool_attrib(ctx, 'static'),
                           pagesize=self.get_expr_attrib(ctx, 'pagesize'))
        form.run(ctx, self.attribs['name'], opts)


tag_classes = (
    FormTag,
)
