#
# Copyright 2011 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#

from albatross.context import NameRecorderMixin, HiddenFieldSessionMixin, \
    PickleSignMixin, CachingTemplateLoaderMixin
from albatross.app import AppContext, Application, PageObjectMixin, \
    PageModuleMixin
from albatross.randompage import RandomPageModuleMixin
from albatross.session import SessionServerContextMixin,\
     SessionServerAppMixin
from albatross.sessionfile import SessionFileContextMixin, \
    SessionFileAppMixin
from albatross.branchingsession import BranchingSessionMixin

# ------------------------------------------------------------------
# Prepackaged application contexts
# ------------------------------------------------------------------

# Simple application context simply stores field names in form to
# allow empty fields to be set to None.  Also saves state in hidden
# form fields.
class SimpleAppContext(AppContext,
                       NameRecorderMixin,
                       HiddenFieldSessionMixin):

    def __init__(self, app):
        AppContext.__init__(self, app)
        NameRecorderMixin.__init__(self)
        HiddenFieldSessionMixin.__init__(self)

    def form_close(self):
        use_multipart_enc = NameRecorderMixin.form_close(self)
        HiddenFieldSessionMixin.form_close(self)
        return use_multipart_enc


# Session state is stored server-side in a session daemon
class SessionAppContext(AppContext,
                        NameRecorderMixin,
                        SessionServerContextMixin):

    def __init__(self, app):
        AppContext.__init__(self, app)
        NameRecorderMixin.__init__(self)
        SessionServerContextMixin.__init__(self)

# Session state is stored server-side in a per-session file
class SessionFileAppContext(AppContext,
                            NameRecorderMixin,
                            SessionFileContextMixin):

    def __init__(self, app):
        AppContext.__init__(self, app)
        NameRecorderMixin.__init__(self)
        SessionFileContextMixin.__init__(self)


# Session state is stored server-side in a session daemon, using hidden fields
# to record session and transaction ID, so that the correct session state is
# used if the user uses the browser <back> button.
class BranchingSessionContext(AppContext,
                              NameRecorderMixin,
                              BranchingSessionMixin):
    def __init__(self, app):
        AppContext.__init__(self, app)
        NameRecorderMixin.__init__(self)
        BranchingSessionMixin.__init__(self)

    def form_close(self):
        use_multipart_enc = NameRecorderMixin.form_close(self)
        BranchingSessionMixin.form_close(self)
        return use_multipart_enc


# ------------------------------------------------------------------
# Prepackaged application objects - SimpleApp
# ------------------------------------------------------------------

# Page processing is implemented in methods of page objects
class SimpleApp(PickleSignMixin,
                Application,
                CachingTemplateLoaderMixin,
                PageObjectMixin):

    def __init__(self, base_url, template_path, start_page, secret):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, template_path)
        PageObjectMixin.__init__(self, start_page)

    def create_context(self):
        return SimpleAppContext(self)


# Same as SimpleApp with server-side session server support
class SimpleSessionApp(PickleSignMixin,
                       Application,
                       CachingTemplateLoaderMixin,
                       PageObjectMixin,
                       SessionServerAppMixin):

    def __init__(self, base_url, template_path, start_page, secret,
                 session_appid, session_server = 'localhost', 
                 server_port = 34343, session_age = 1800):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, template_path)
        PageObjectMixin.__init__(self, start_page)
        SessionServerAppMixin.__init__(self, session_appid, session_server, 
                                       server_port, session_age)

    def create_context(self):
        return SessionAppContext(self)


class BranchingSimpleSessionApp(SimpleSessionApp):

    def create_context(self):
        return BranchingSessionContext(self)


# Same as SimpleApp with server-side session file support
class SimpleSessionFileApp(PickleSignMixin,
                            Application,
                            CachingTemplateLoaderMixin,
                            PageObjectMixin,
                            SessionFileAppMixin):

    def __init__(self, base_url, template_path, start_page, secret,
                 session_appid, session_dir):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, template_path)
        PageObjectMixin.__init__(self, start_page)
        SessionFileAppMixin.__init__(self, session_appid, session_dir)

    def create_context(self):
        return SessionFileAppContext(self)


# ------------------------------------------------------------------
# Prepackaged application objects - ModularApp
# ------------------------------------------------------------------

# Page processing is implemented in functions of page modules
class ModularApp(PickleSignMixin,
                 Application,
                 CachingTemplateLoaderMixin,
                 PageModuleMixin):

    def __init__(self, base_url, module_path, template_path, start_page, secret):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, template_path)
        PageModuleMixin.__init__(self, module_path, start_page)

    def create_context(self):
        return SimpleAppContext(self)


# Same as ModularApp with server-side session server support
class ModularSessionApp(PickleSignMixin,
                        Application,
                        CachingTemplateLoaderMixin,
                        PageModuleMixin,
                        SessionServerAppMixin):

    def __init__(self, base_url, module_path, template_path, start_page, secret,
                 session_appid, session_server = 'localhost', 
                 server_port = 34343, session_age = 1800):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, template_path)
        PageModuleMixin.__init__(self, module_path, start_page)
        SessionServerAppMixin.__init__(self, session_appid, session_server, 
                                       server_port, session_age)

    def create_context(self):
        return SessionAppContext(self)


class BranchingModularSessionApp(ModularSessionApp):

    def create_context(self):
        return BranchingSessionContext(self)


# Same as ModularApp with server-side session file support
class ModularSessionFileApp(PickleSignMixin,
                            Application,
                            CachingTemplateLoaderMixin,
                            PageModuleMixin,
                            SessionFileAppMixin):

    def __init__(self, base_url, module_path, template_path, start_page, secret,
                 session_appid, session_dir):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, template_path)
        PageModuleMixin.__init__(self, module_path, start_page)
        SessionFileAppMixin.__init__(self, session_appid, session_dir)

    def create_context(self):
        return SessionFileAppContext(self)


# ------------------------------------------------------------------
# Prepackaged application objects - RandomApp
# ------------------------------------------------------------------

# URI specifies page with hidden field session support
# NOTE - this combination DOES NOT WORK, as the hidden field session is lost
# when the application is accessed randomly.
class RandomModularApp(PickleSignMixin,
                       Application,
                       CachingTemplateLoaderMixin,
                       RandomPageModuleMixin):
    def __init__(self, base_url, page_path, start_page, secret):
        PickleSignMixin.__init__(self, secret)
        Application.__init__(self, base_url)
        CachingTemplateLoaderMixin.__init__(self, page_path)
        RandomPageModuleMixin.__init__(self, page_path, start_page)

    def create_context(self):
        return SimpleAppContext(self)


# URI specifies page with server-side session server support
class RandomModularSessionApp(PickleSignMixin,
                              Application,
                              CachingTemplateLoaderMixin,
                              RandomPageModuleMixin,
                              SessionServerAppMixin):

    def __init__(self, base_url, page_path, start_page, secret,
                 session_appid, session_server = 'localhost', server_port = 34343, session_age = 1800):
        PickleSignMixin.__init__(self, secret)
        Application.__init__(self, base_url)
        CachingTemplateLoaderMixin.__init__(self, page_path)
        RandomPageModuleMixin.__init__(self, page_path, start_page)
        SessionServerAppMixin.__init__(self, session_appid, session_server, server_port, session_age)

    def create_context(self):
        return SessionAppContext(self)


# URI specifies page with server-side session file support
class RandomModularSessionFileApp(PickleSignMixin,
                                  Application,
                                  CachingTemplateLoaderMixin,
                                  RandomPageModuleMixin,
                                  SessionFileAppMixin):

    def __init__(self, base_url, page_path, start_page, secret,
                 session_appid, session_dir):
        Application.__init__(self, base_url)
        PickleSignMixin.__init__(self, secret)
        CachingTemplateLoaderMixin.__init__(self, page_path)
        RandomPageModuleMixin.__init__(self, page_path, start_page)
        SessionFileAppMixin.__init__(self, session_appid, session_dir)

    def create_context(self):
        return SessionFileAppContext(self)


# ------------------------------------------------------------------
# Pre-packaged Application factory function
# ------------------------------------------------------------------

app_classes = {
    'simple': {
        'hidden-field': 
            (SimpleApp, SimpleAppContext),
        'session-server':
            (SimpleSessionApp, SessionAppContext),
        'session-file':
            (SimpleSessionFileApp, SessionFileAppContext),
        'branching-server':
            (BranchingSimpleSessionApp, BranchingSessionContext),
    },
    'modular': {
        'hidden-field':
            (ModularApp, SimpleAppContext),
        'session-server':
            (ModularSessionApp, SessionAppContext),
        'session-file':
            (ModularSessionFileApp, SessionFileAppContext),
        'branching-server':
            (BranchingModularSessionApp, BranchingSessionContext),
    },
    'random': {
        'session-server':
            (RandomModularSessionApp, SessionAppContext),
        'session-file':
            (RandomModularSessionFileApp, SessionFileAppContext),
# Random access not compatible with hidden field sessions
#        'hidden-field':
#            (RandomModularApp, SimpleAppContext),
#        'branching-server':
#            (BranchingRandomModularSessionApp, BranchingSessionContext),
    },
}


def get_app_classes(page_model, context_model):
    try:
        context_classes = app_classes[page_model]
    except KeyError:
        raise ApplicationError('Unknown page model: %r' % page_model)
    try:
        return context_classes[context_model]
    except KeyError:
        raise ApplicationError('Unknown context model: %r' % context_model)
