try:
    import sqlite3 as sqlite
    SUBST = '?'
except ImportError:
    import sqlite
    SUBST = '%s'


(ORDER_DESC,
 ORDER_ASC) = range(2)

ORDER = {ORDER_DESC: 'DESC',
         ORDER_ASC: 'ASC'}


class DatabaseError(Exception):
    pass


class Row:

    table = None

    def __init__(self, **attrs):
        self.__cols = attrs.keys()
        self.__dict__.update(attrs)

    def column_names(self):
        return self.__cols

    def _db_attrs(self):
        return dict([(col, getattr(self, col))
                     for col in self.__cols])

    def delete(self, conn):
        return conn.delete(self.table, id=self.id)

    def save(self, conn):
        if hasattr(self, 'id') and self.id:
            set_args = self._db_attrs()
            where_args = {'id': set_args.pop('id')}
            return conn.update(self.table, set_args, where_args)
        else:
            return self.insert(conn)

    def insert(self, conn):
        cursor = conn.insert(self.table, **self._db_attrs())
        self.id = cursor.lastrowid
        if 'id' not in self.__cols:
            self.__cols.append('id')
        return cursor


class Database:

    def __init__(self, path):
        try:
            # old sqlite defaults to 'ascii' and returns TEXT as a
            # byte-string
            self.db = sqlite.connect(path,
                                     isolation_level=None,
                                     encoding='utf-8')
        except TypeError:
            # sqlite3 is unicode ('encoding' keyword arg removed) and
            # returns TEXT as unicode unless an alternative
            # text_factory is specified.
            self.db = sqlite.connect(path,
                                     isolation_level=None)
            self.db.text_factory = str

    def commit(self):
        self.db.commit()

    def fetch_empty_row(self, row_class):
        sql = 'SELECT * FROM %s WHERE 0' % row_class.table
        cursor = self.db.cursor()
        cursor.execute(sql)
        attrs = dict([(col[0], None)
                      for col in cursor.description])
        return row_class(**attrs)

    def fetch_all(self, table=None, row_class=Row, order_args=None,
                  offset=None, limit=None, **kwargs):
        if table is None and row_class.table:
            table = row_class.table
        sql = 'SELECT * FROM %s' % table
        args = []
        where = []
        for key, value in kwargs.items():
            where.append('%s = %s' % (key, SUBST))
            args.append(value)
        if where:
            sql = '%s WHERE %s' % (sql, ' AND '.join(where))
        order = []
        if order_args:
            for col, direction in order_args:
                order.append('%s %s' % (col, ORDER[direction]))
        if order:
            sql = '%s ORDER BY %s' % (sql, ' , '.join(order))
        cursor = self.db.cursor()
        if limit:
            sql = '%s LIMIT %s' % (sql, limit)
        if offset:
            sql = '%s OFFSET %s' % (sql, offset)
        cursor.execute(sql, args)
        rows = []
        for row in cursor.fetchall():
            attrs = dict([(col, value)
                          for col, value in zip(
                        [col[0] for col in cursor.description], row)])
            r = row_class(**attrs)
            if r.table != table:
                r.table = table
            rows.append(r)
        return rows

    def fetch_one(self, table=None, row_class=Row, **kwargs):
        rows = self.fetch_all(table=table, row_class=row_class, **kwargs)
        if not rows:
            raise DatabaseError('No rows for query - expected one.')
        if len(rows) > 1:
            raise DatabaseError('Multiple rows for query - expected one.')
        return rows[0]

    def count(self, table):
        sql = 'SELECT COUNT(*) FROM %s' % table
        cursor = self.db.cursor()
        cursor.execute(sql)
        return cursor.fetchall()[0][0]

    def insert(self, table, **kwargs):
        cols = []
        values = []
        for col, value in kwargs.items():
            cols.append(col)
            values.append(value)
        sql = 'INSERT INTO %s (%s) VALUES (%s)' \
            % (table, ', '.join(cols), ', '.join([SUBST] * len(cols)))
        cursor = self.db.cursor()
        cursor.execute(sql, values)
        return cursor

    def delete(self, table, **kwargs):
        sql = 'DELETE from %s' % table
        args = []
        where = []
        for key, value in kwargs.items():
            where.append('%s = %s' % (key, SUBST))
            args.append(value)
        if where:
            sql = '%s WHERE %s' % (sql, ' AND '.join(where))
        cursor = self.db.cursor()
        cursor.execute(sql, args)
        return cursor

    def update(self, table, set_args, where_args):
        sql = 'UPDATE %s' % table
        args = []
        set = []
        where = []
        for key, value in set_args.items():
            set.append('%s = %s' % (key, SUBST))
            args.append(value)
        for key, value in where_args.items():
            where.append('%s = %s' % (key, SUBST))
            args.append(value)
        if set:
            sql = '%s SET %s' % (sql, ', '.join(set))
        if where:
            sql = '%s WHERE %s' % (sql, ' AND '.join(where))
        cursor = self.db.cursor()
        cursor.execute(sql, args)
        return cursor
